@@ -95,9 +95,11 @@ void run_instance
 #ifdef DKGPG_TESTSUITE
 		passphrase = "Test";
 #else
-		std::cout << "Please enter the passphrase to unlock your private key: ";
-		std::getline(std::cin, passphrase);
-		std::cin.clear();
+		if (!get_passphrase("Please enter passphrase to unlock your private key", passphrase))
+		{
+			release_mpis();
+			exit(-1);
+		}
 #endif
 		if (!parse_private_key(armored_seckey, ckeytime, ekeytime, CAPL))
 		{
@@ -297,24 +299,21 @@ void run_instance
 	// sign the hashes
 	tmcg_octets_t revsig_pub, revsig_sub;
 	tmcg_byte_t buffer[1024];
-	gcry_mpi_t h, r, s;
+	gcry_mpi_t r, s, h;
 	mpz_t dsa_m, dsa_r, dsa_s;
 	size_t buflen = 0;
 	gcry_error_t ret;
-	h = gcry_mpi_new(2048);
-	r = gcry_mpi_new(2048);
-	s = gcry_mpi_new(2048);
-	mpz_init(dsa_m), mpz_init(dsa_r), mpz_init(dsa_s);
 	memset(buffer, 0, sizeof(buffer));
 	for (size_t i = 0; ((i < hash_pub.size()) && (i < sizeof(buffer))); i++, buflen++)
 		buffer[i] = hash_pub[i];
+	r = gcry_mpi_new(2048);
+	s = gcry_mpi_new(2048);
+	mpz_init(dsa_m), mpz_init(dsa_r), mpz_init(dsa_s);
 	ret = gcry_mpi_scan(&h, GCRYMPI_FMT_USG, buffer, buflen, NULL);
 	if (ret)
 	{
 		std::cerr << "R_" << whoami << ": gcry_mpi_scan() failed for h" << std::endl;
-		gcry_mpi_release(h);
-		gcry_mpi_release(r);
-		gcry_mpi_release(s);
+		gcry_mpi_release(r), gcry_mpi_release(s);
 		mpz_clear(dsa_m), mpz_clear(dsa_r), mpz_clear(dsa_s);
 		delete dss, delete rbc, delete aiou, delete aiou2;
 		release_mpis();
@@ -323,14 +322,13 @@ void run_instance
 	if (!mpz_set_gcry_mpi(h, dsa_m))
 	{
 		std::cerr << "R_" << whoami << ": mpz_set_gcry_mpi() failed for dsa_m" << std::endl;
-		gcry_mpi_release(h);
-		gcry_mpi_release(r);
-		gcry_mpi_release(s);
+		gcry_mpi_release(r), gcry_mpi_release(s), gcry_mpi_release(h);
 		mpz_clear(dsa_m), mpz_clear(dsa_r), mpz_clear(dsa_s);
 		delete dss, delete rbc, delete aiou, delete aiou2;
 		release_mpis();
 		exit(-1);
 	}
+	gcry_mpi_release(h);
 	std::stringstream err_log_sign;
 	if (opt_verbose)
 		std::cout << "R_" << whoami << ": dss.Sign() on pub" << std::endl;
@@ -338,9 +336,7 @@ void run_instance
 	{
 		std::cerr << "R_" << whoami << ": " << "tDSS Sign() on pub failed" << std::endl;
 		std::cerr << "R_" << whoami << ": log follows " << std::endl << err_log_sign.str();
-		gcry_mpi_release(h);
-		gcry_mpi_release(r);
-		gcry_mpi_release(s);
+		gcry_mpi_release(r), gcry_mpi_release(s);
 		mpz_clear(dsa_m), mpz_clear(dsa_r), mpz_clear(dsa_s);
 		delete dss, delete rbc, delete aiou, delete aiou2;
 		release_mpis();
@@ -348,23 +344,19 @@ void run_instance
 	}
 	if (opt_verbose > 1)
 		std::cout << "R_" << whoami << ": log follows " << std::endl << err_log_sign.str();
-	if (!mpz_get_gcry_mpi(&r, dsa_r))
+	if (!mpz_get_gcry_mpi(r, dsa_r))
 	{
 		std::cerr << "R_" << whoami << ": mpz_get_gcry_mpi() failed for dsa_r" << std::endl;
-		gcry_mpi_release(h);
-		gcry_mpi_release(r);
-		gcry_mpi_release(s);
+		gcry_mpi_release(r), gcry_mpi_release(s);
 		mpz_clear(dsa_m), mpz_clear(dsa_r), mpz_clear(dsa_s);
 		delete dss, delete rbc, delete aiou, delete aiou2;
 		release_mpis();
 		exit(-1);
 	}
-	if (!mpz_get_gcry_mpi(&s, dsa_s))
+	if (!mpz_get_gcry_mpi(s, dsa_s))
 	{
 		std::cerr << "R_" << whoami << ": mpz_get_gcry_mpi() failed for dsa_s" << std::endl;
-		gcry_mpi_release(h);
-		gcry_mpi_release(r);
-		gcry_mpi_release(s);
+		gcry_mpi_release(r), gcry_mpi_release(s);
 		mpz_clear(dsa_m), mpz_clear(dsa_r), mpz_clear(dsa_s);
 		delete dss, delete rbc, delete aiou, delete aiou2;
 		release_mpis();
@@ -379,9 +371,7 @@ void run_instance
 	if (ret)
 	{
 		std::cerr << "R_" << whoami << ": gcry_mpi_scan() failed for h" << std::endl;
-		gcry_mpi_release(h);
-		gcry_mpi_release(r);
-		gcry_mpi_release(s);
+		gcry_mpi_release(r), gcry_mpi_release(s);
 		mpz_clear(dsa_m), mpz_clear(dsa_r), mpz_clear(dsa_s);
 		delete dss, delete rbc, delete aiou, delete aiou2;
 		release_mpis();
@@ -390,14 +380,13 @@ void run_instance
 	if (!mpz_set_gcry_mpi(h, dsa_m))
 	{
 		std::cerr << "R_" << whoami << ": mpz_set_gcry_mpi() failed for dsa_m" << std::endl;
-		gcry_mpi_release(h);
-		gcry_mpi_release(r);
-		gcry_mpi_release(s);
+		gcry_mpi_release(r), gcry_mpi_release(s), gcry_mpi_release(h);
 		mpz_clear(dsa_m), mpz_clear(dsa_r), mpz_clear(dsa_s);
 		delete dss, delete rbc, delete aiou, delete aiou2;
 		release_mpis();
 		exit(-1);
 	}
+	gcry_mpi_release(h);
 	std::stringstream err_log_sign2;
 	if (opt_verbose)
 		std::cout << "R_" << whoami << ": dss.Sign() on sub" << std::endl;
@@ -405,9 +394,7 @@ void run_instance
 	{
 		std::cerr << "R_" << whoami << ": " << "tDSS Sign() on sub failed" << std::endl;
 		std::cerr << "R_" << whoami << ": log follows " << std::endl << err_log_sign2.str();
-		gcry_mpi_release(h);
-		gcry_mpi_release(r);
-		gcry_mpi_release(s);
+		gcry_mpi_release(r), gcry_mpi_release(s);
 		mpz_clear(dsa_m), mpz_clear(dsa_r), mpz_clear(dsa_s);
 		delete dss, delete rbc, delete aiou, delete aiou2;
 		release_mpis();
@@ -415,32 +402,26 @@ void run_instance
 	}
 	if (opt_verbose > 1)
 		std::cout << "R_" << whoami << ": log follows " << std::endl << err_log_sign2.str();
-	if (!mpz_get_gcry_mpi(&r, dsa_r))
+	if (!mpz_get_gcry_mpi(r, dsa_r))
 	{
 		std::cerr << "R_" << whoami << ": mpz_get_gcry_mpi() failed for dsa_r" << std::endl;
-		gcry_mpi_release(h);
-		gcry_mpi_release(r);
-		gcry_mpi_release(s);
+		gcry_mpi_release(r), gcry_mpi_release(s);
 		mpz_clear(dsa_m), mpz_clear(dsa_r), mpz_clear(dsa_s);
 		delete dss, delete rbc, delete aiou, delete aiou2;
 		release_mpis();
 		exit(-1);
 	}
-	if (!mpz_get_gcry_mpi(&s, dsa_s))
+	if (!mpz_get_gcry_mpi(s, dsa_s))
 	{
 		std::cerr << "R_" << whoami << ": mpz_get_gcry_mpi() failed for dsa_s" << std::endl;
-		gcry_mpi_release(h);
-		gcry_mpi_release(r);
-		gcry_mpi_release(s);
+		gcry_mpi_release(r), gcry_mpi_release(s);
 		mpz_clear(dsa_m), mpz_clear(dsa_r), mpz_clear(dsa_s);
 		delete dss, delete rbc, delete aiou, delete aiou2;
 		release_mpis();
 		exit(-1);
 	}
 	CallasDonnerhackeFinneyShawThayerRFC4880::PacketSigEncode(trailer_sub, left_sub, r, s, revsig_sub);
-	gcry_mpi_release(h);
-	gcry_mpi_release(r);
-	gcry_mpi_release(s);
+	gcry_mpi_release(r), gcry_mpi_release(s);
 	mpz_clear(dsa_m), mpz_clear(dsa_r), mpz_clear(dsa_s);
 
 	// at the end: deliver some more rounds for still waiting parties
@@ -712,11 +693,19 @@ int main
 	opt_r = 3;
 	opt_verbose = 1;
 #endif
+
+	// check command line arguments
+	if ((opt_hostname != NULL) && (opt_passwords == NULL))
+	{
+		std::cerr << "ERROR: option \"-P\" is necessary due to insecure network" << std::endl;
+		return -1;
+	}
 	if (peers.size() < 1)
 	{
 		std::cerr << "ERROR: no peers given as argument; usage: " << usage << std::endl;
 		return -1;
 	}
+
 	// canonicalize peer list
 	std::sort(peers.begin(), peers.end());
 	std::vector<std::string>::iterator it = std::unique(peers.begin(), peers.end());
@@ -726,22 +715,22 @@ int main
 		std::cerr << "ERROR: too few or too many peers given" << std::endl;
 		return -1;
 	}
-	if (!init_libTMCG())
-	{
-		std::cerr << "ERROR: initialization of LibTMCG failed" << std::endl;
-		return -1;
-	}
-	if ((opt_hostname != NULL) && (opt_passwords == NULL))
-	{
-		std::cerr << "ERROR: option \"-P\" is necessary due to insecure network" << std::endl;
-		return -1;
-	}
 	if (opt_verbose)
 	{
 		std::cout << "INFO: canonicalized peer list = " << std::endl;
 		for (size_t i = 0; i < peers.size(); i++)
 			std::cout << peers[i] << std::endl;
 	}
+
+	// initialize LibTMCG
+	if (!init_libTMCG())
+	{
+		std::cerr << "ERROR: initialization of LibTMCG failed" << std::endl;
+		return -1;
+	}
+	if (opt_verbose)
+		std::cout << "INFO: using LibTMCG version " << version_libTMCG() << std::endl;
+	
 	// initialize return code
 	int ret = 0;
 	// create underlying point-to-point channels, if built-in TCP/IP service requested
