@@ -75,7 +75,7 @@ char					*opt_ifilename = NULL;
 char					*opt_ofilename = NULL;
 char					*opt_passwords = NULL;
 char					*opt_hostname = NULL;
-unsigned long int			opt_p = 55000;
+unsigned long int			opt_p = 55000, opt_W = 5;
 
 std::string				armored_message;
 
@@ -467,8 +467,7 @@ bool combine_decryption_shares
 		}
 
 		// copy the result from R to gk
-		gcry_mpi_release(gk);
-		if (!mpz_get_gcry_mpi(&gk, R))
+		if (!mpz_get_gcry_mpi(gk, R))
 		{
 			std::cerr << "ERROR: converting interpolated result failed" << std::endl;
 			exit(-1);
@@ -535,9 +534,11 @@ void run_instance
 #ifdef DKGPG_TESTSUITE
 		passphrase = "Test";
 #else
-		std::cout << "Please enter the passphrase to unlock your private key: ";
-		std::getline(std::cin, passphrase);
-		std::cin.clear();
+		if (!get_passphrase("Please enter passphrase to unlock your private key", passphrase))
+		{
+			release_mpis();
+			exit(-1);
+		}
 #endif
 		if (!parse_private_key(armored_seckey, ckeytime, ekeytime, CAPL))
 		{
@@ -593,10 +594,10 @@ void run_instance
 	}
 
 	// create asynchronous authenticated unicast channels
-	aiounicast_select *aiou = new aiounicast_select(peers.size(), whoami, uP_in, uP_out, uP_key);
+	aiounicast_select *aiou = new aiounicast_select(peers.size(), whoami, uP_in, uP_out, uP_key, aiounicast::aio_scheduler_roundrobin, (opt_W * 60));
 
 	// create asynchronous authenticated unicast channels for broadcast protocol
-	aiounicast_select *aiou2 = new aiounicast_select(peers.size(), whoami, bP_in, bP_out, bP_key);
+	aiounicast_select *aiou2 = new aiounicast_select(peers.size(), whoami, bP_in, bP_out, bP_key, aiounicast::aio_scheduler_roundrobin, (opt_W * 60));
 			
 	// create an instance of a reliable broadcast protocol (RBC)
 	std::string myID = "dkg-decrypt|";
@@ -604,7 +605,7 @@ void run_instance
 		myID += peers[i] + "|";
 	myID += dkg->t; // include parameterized t-resiliance of DKG in the ID of broadcast protocol
 	size_t T_RBC = (peers.size() - 1) / 3; // assume maximum asynchronous t-resilience for RBC
-	CachinKursawePetzoldShoupRBC *rbc = new CachinKursawePetzoldShoupRBC(peers.size(), T_RBC, whoami, aiou2);
+	CachinKursawePetzoldShoupRBC *rbc = new CachinKursawePetzoldShoupRBC(peers.size(), T_RBC, whoami, aiou2, aiounicast::aio_scheduler_roundrobin, (opt_W * 60));
 	rbc->setID(myID);
 
 	// perform a simple exchange test with debug output
@@ -892,6 +893,7 @@ char *gnunet_opt_passwords = NULL;
 char *gnunet_opt_port = NULL;
 unsigned int gnunet_opt_xtests = 0;
 unsigned int gnunet_opt_wait = 5;
+unsigned int gnunet_opt_W = opt_W;
 int gnunet_opt_nonint = 0;
 int gnunet_opt_verbose = 0;
 #endif
@@ -985,6 +987,12 @@ int main
 			"minutes to wait until start of decryption",
 			&gnunet_opt_wait
 		),
+		GNUNET_GETOPT_option_uint('W',
+			"aiou-timeout",
+			"TIME",
+			"timeout for point-to-point messages in minutes",
+			&gnunet_opt_W
+		),
 		GNUNET_GETOPT_option_uint('x',
 			"x-tests",
 			NULL,
@@ -1015,6 +1023,8 @@ int main
 		passwords = gnunet_opt_passwords; // get passwords from GNUnet options
 	if (gnunet_opt_hostname != NULL)
 		hostname = gnunet_opt_hostname; // get hostname from GNUnet options
+	if (gnunet_opt_W != opt_W)
+		opt_W = gnunet_opt_W; // get aiou message timeout from GNUnet options
 #endif
 
 	bool nonint = false;
@@ -1024,7 +1034,8 @@ int main
 		std::string arg = argv[i+1];
 		// ignore options
 		if ((arg.find("-c") == 0) || (arg.find("-p") == 0) || (arg.find("-w") == 0) || (arg.find("-L") == 0) || (arg.find("-l") == 0) ||
-			(arg.find("-i") == 0) || (arg.find("-o") == 0) || (arg.find("-x") == 0) || (arg.find("-P") == 0) || (arg.find("-H") == 0))
+			(arg.find("-i") == 0) || (arg.find("-o") == 0) || (arg.find("-x") == 0) || (arg.find("-P") == 0) || (arg.find("-H") == 0) ||
+			(arg.find("-W") == 0))
 		{
 			size_t idx = ++i;
 			if ((arg.find("-i") == 0) && (idx < (size_t)(argc - 1)) && (opt_ifilename == NULL))
@@ -1049,6 +1060,8 @@ int main
 			}
 			if ((arg.find("-p") == 0) && (idx < (size_t)(argc - 1)) && (port.length() == 0))
 				port = argv[i+1];
+			if ((arg.find("-W") == 0) && (idx < (size_t)(argc - 1)) && (opt_W == 5))
+				opt_W = strtoul(argv[i+1], NULL, 10);
 			continue;
 		}
 		else if ((arg.find("--") == 0) || (arg.find("-v") == 0) || (arg.find("-h") == 0) || (arg.find("-n") == 0) || (arg.find("-V") == 0))
@@ -1068,6 +1081,7 @@ int main
 				std::cout << "  -P STRING              exchanged passwords to protect private and broadcast channels" << std::endl;
 				std::cout << "  -v, --version          print the version number" << std::endl;
 				std::cout << "  -V, --verbose          turn on verbose output" << std::endl;
+				std::cout << "  -W TIME                timeout for point-to-point messages in minutes" << std::endl;
 #endif
 				return 0; // not continue
 			}
@@ -1108,11 +1122,19 @@ int main
 	opt_ifilename = (char*)ifilename.c_str();
 	opt_verbose = 1;
 #endif
+
+	// check command line arguments
+	if ((opt_hostname != NULL) && (opt_passwords == NULL))
+	{
+		std::cerr << "ERROR: option \"-P\" is necessary due to insecure network" << std::endl;
+		return -1;
+	}
 	if (peers.size() < 1)
 	{
 		std::cerr << "ERROR: no peers given as argument; usage: " << usage << std::endl;
 		return -1;
 	}
+
 	// canonicalize peer list
 	std::sort(peers.begin(), peers.end());
 	std::vector<std::string>::iterator it = std::unique(peers.begin(), peers.end());
@@ -1127,16 +1149,23 @@ int main
 		std::cerr << "ERROR: too few or too many peers given" << std::endl;
 		return -1;
 	}
+	if (opt_verbose)
+	{
+		std::cout << "INFO: canonicalized peer list = " << std::endl;
+		for (size_t i = 0; i < peers.size(); i++)
+			std::cout << peers[i] << std::endl;
+	}
+
+	// initialize LibTMCG
 	if (!init_libTMCG())
 	{
 		std::cerr << "ERROR: initialization of LibTMCG failed" << std::endl;
 		return -1;
 	}
-	if ((opt_hostname != NULL) && (opt_passwords == NULL))
-	{
-		std::cerr << "ERROR: option \"-P\" is necessary due to insecure network" << std::endl;
-		return -1;
-	}
+	if (opt_verbose)
+		std::cout << "INFO: using LibTMCG version " << version_libTMCG() << std::endl;
+
+	// read message
 	if (opt_ifilename != NULL)
 	{
 		if (!read_message(opt_ifilename, armored_message))
@@ -1150,12 +1179,8 @@ int main
 			armored_message += line + "\r\n";
 		std::cin.clear();
 	}
-	if (opt_verbose)
-	{
-		std::cout << "INFO: canonicalized peer list = " << std::endl;
-		for (size_t i = 0; i < peers.size(); i++)
-			std::cout << peers[i] << std::endl;
-	}
+
+	// start non-interactive variant
 	if (nonint)
 	{
 		size_t idx;
@@ -1177,9 +1202,11 @@ int main
 			dss_qual.clear(), dss_x_rvss_qual.clear(), dss_c_ik.clear(), dkg_qual.clear(), dkg_v_i.clear(), dkg_c_ik.clear();
 			init_mpis();
 			// protected with password
-			std::cout << "Please enter the passphrase to unlock your private key: ";
-			std::getline(std::cin, passphrase);
-			std::cin.clear();
+			if (!get_passphrase("Please enter passphrase to unlock your private key", passphrase))
+			{
+				release_mpis();
+				return -1;
+			}
 			if (!parse_private_key(armored_seckey, ckeytime, ekeytime, CAPL))
 			{
 				std::cerr << "ERROR: wrong passphrase to unlock private key" << std::endl;
@@ -1348,6 +1375,12 @@ int main
 			"minutes to wait until start of decryption",
 			&gnunet_opt_wait
 		),
+		GNUNET_GETOPT_option_uint('W',
+			"aiou-timeout",
+			"TIME",
+			"timeout for point-to-point messages in minutes",
+			&gnunet_opt_W
+		),
 		GNUNET_GETOPT_option_uint('x',
 			"x-tests",
 			NULL,
@@ -1363,7 +1396,7 @@ int main
 	else
 		return -1;
 #else
-	std::cerr << "WARNING: GNUnet development files are required for message exchange of decryption protocol" << std::endl;
+	std::cerr << "WARNING: GNUnet CADET is required for the message exchange of this program" << std::endl;
 #endif
 
 	std::cout << "INFO: running local test with " << peers.size() << " participants" << std::endl;
