@@ -44,27 +44,42 @@ static const char *about = PACKAGE_STRING " " PACKAGE_URL;
 #include <libTMCG.hh>
 #include <aiounicast_select.hh>
 
+#include "dkg-common.hh"
 #include "dkg-tcpip-common.hh"
 #include "dkg-gnunet-common.hh"
 
-int				pipefd[DKGPG_MAX_N][DKGPG_MAX_N][2], broadcast_pipefd[DKGPG_MAX_N][DKGPG_MAX_N][2];
-pid_t				pid[DKGPG_MAX_N];
-std::vector<std::string>	peers;
-bool				instance_forked = false;
+int					pipefd[DKGPG_MAX_N][DKGPG_MAX_N][2], broadcast_pipefd[DKGPG_MAX_N][DKGPG_MAX_N][2];
+pid_t					pid[DKGPG_MAX_N];
+std::vector<std::string>		peers;
+bool					instance_forked = false;
 
-std::string			userid, passphrase, passwords, hostname, port;
-int 				opt_verbose = 0;
-char				*opt_crs = NULL;
-char				*opt_passwords = NULL;
-char				*opt_hostname = NULL;
-unsigned long int		opt_t = DKGPG_MAX_N, opt_s = DKGPG_MAX_N, opt_e = 0, opt_p = 55000, opt_W = 5;
+std::string				passphrase, userid, passwords, hostname, port;
+tmcg_octets_t				keyid, subkeyid, pub, sub, uidsig, subsig, sec, ssb, uid;
+std::map<size_t, size_t>		idx2dkg, dkg2idx;
+mpz_t					dss_p, dss_q, dss_g, dss_h, dss_x_i, dss_xprime_i, dss_y;
+size_t					dss_n, dss_t, dss_i;
+std::vector<size_t>			dss_qual, dss_x_rvss_qual;
+std::vector< std::vector<mpz_ptr> >	dss_c_ik;
+mpz_t					dkg_p, dkg_q, dkg_g, dkg_h, dkg_x_i, dkg_xprime_i, dkg_y;
+size_t					dkg_n, dkg_t, dkg_i;
+std::vector<size_t>			dkg_qual;
+std::vector<mpz_ptr>			dkg_v_i;
+std::vector< std::vector<mpz_ptr> >	dkg_c_ik;
+gcry_mpi_t 				dsa_p, dsa_q, dsa_g, dsa_y, dsa_x, elg_p, elg_q, elg_g, elg_y, elg_x;
+gcry_mpi_t 				gk, myk, sig_r, sig_s;
 
-bool				fips = false;
-std::stringstream		crss;
-mpz_t 				cache[TMCG_MAX_SSRANDOMM_CACHE], cache_mod;
-size_t				cache_avail = 0;
+int 					opt_verbose = 0;
+char					*opt_crs = NULL;
+char					*opt_passwords = NULL;
+char					*opt_hostname = NULL;
+unsigned long int			opt_t = DKGPG_MAX_N, opt_s = DKGPG_MAX_N, opt_e = 0, opt_p = 55000, opt_W = 5;
 
-size_t				T, S;
+bool					fips = false;
+std::stringstream			crss;
+mpz_t 					cache[TMCG_MAX_SSRANDOMM_CACHE], cache_mod;
+size_t					cache_avail = 0;
+
+size_t					T, S;
 
 void run_instance
 	(const size_t whoami, const time_t keytime, const time_t keyexptime, const size_t num_xtests)
@@ -341,33 +356,29 @@ void run_instance
 		mpz_ssrandomm_cache(cache, cache_mod, &cache_avail, dsa_x, vtmf->q); // choose private key for DSA
 		mpz_spowm(dsa_y, vtmf->g, dsa_x, vtmf->p); // compute public key for DSA
 	}
-	if (!mpz_get_gcry_mpi(&p, vtmf->p))
+	p = gcry_mpi_new(2048);
+	if (!mpz_get_gcry_mpi(p, vtmf->p))
 	{
 		std::cerr << "P_" << whoami << ": mpz_get_gcry_mpi() failed for p" << std::endl;
 		mpz_clear(dsa_y), mpz_clear(dsa_x), mpz_clear(dsa_m), mpz_clear(dsa_r), mpz_clear(dsa_s);
-		delete dkg, delete dss, delete rbc, delete vtmf, delete aiou, delete aiou2;
-		exit(-1);
-	}
-	if (!mpz_get_gcry_mpi(&q, vtmf->q))
-	{
-		std::cerr << "P_" << whoami << ": mpz_get_gcry_mpi() failed for q" << std::endl;
-		mpz_clear(dsa_y), mpz_clear(dsa_x), mpz_clear(dsa_m), mpz_clear(dsa_r), mpz_clear(dsa_s);
 		gcry_mpi_release(p);
 		delete dkg, delete dss, delete rbc, delete vtmf, delete aiou, delete aiou2;
 		exit(-1);
 	}
-	if (!mpz_get_gcry_mpi(&g, vtmf->g))
+	q = gcry_mpi_new(2048);
+	if (!mpz_get_gcry_mpi(q, vtmf->q))
 	{
-		std::cerr << "P_" << whoami << ": mpz_get_gcry_mpi() failed for g" << std::endl;
+		std::cerr << "P_" << whoami << ": mpz_get_gcry_mpi() failed for q" << std::endl;
 		mpz_clear(dsa_y), mpz_clear(dsa_x), mpz_clear(dsa_m), mpz_clear(dsa_r), mpz_clear(dsa_s);
 		gcry_mpi_release(p);
 		gcry_mpi_release(q);
 		delete dkg, delete dss, delete rbc, delete vtmf, delete aiou, delete aiou2;
 		exit(-1);
 	}
-	if (!mpz_get_gcry_mpi(&y, dsa_y))
+	g = gcry_mpi_new(2048);
+	if (!mpz_get_gcry_mpi(g, vtmf->g))
 	{
-		std::cerr << "P_" << whoami << ": mpz_get_gcry_mpi() failed for dsa_y" << std::endl;
+		std::cerr << "P_" << whoami << ": mpz_get_gcry_mpi() failed for g" << std::endl;
 		mpz_clear(dsa_y), mpz_clear(dsa_x), mpz_clear(dsa_m), mpz_clear(dsa_r), mpz_clear(dsa_s);
 		gcry_mpi_release(p);
 		gcry_mpi_release(q);
@@ -375,7 +386,20 @@ void run_instance
 		delete dkg, delete dss, delete rbc, delete vtmf, delete aiou, delete aiou2;
 		exit(-1);
 	}
-	if (!mpz_get_gcry_mpi(&x, dsa_x))
+	y = gcry_mpi_new(2048);
+	if (!mpz_get_gcry_mpi(y, dsa_y))
+	{
+		std::cerr << "P_" << whoami << ": mpz_get_gcry_mpi() failed for dsa_y" << std::endl;
+		mpz_clear(dsa_y), mpz_clear(dsa_x), mpz_clear(dsa_m), mpz_clear(dsa_r), mpz_clear(dsa_s);
+		gcry_mpi_release(p);
+		gcry_mpi_release(q);
+		gcry_mpi_release(g);
+		gcry_mpi_release(y);
+		delete dkg, delete dss, delete rbc, delete vtmf, delete aiou, delete aiou2;
+		exit(-1);
+	}
+	x = gcry_mpi_new(2048);
+	if (!mpz_get_gcry_mpi(x, dsa_x))
 	{
 		std::cerr << "P_" << whoami << ": mpz_get_gcry_mpi() failed for dsa_x" << std::endl;
 		mpz_clear(dsa_y), mpz_clear(dsa_x), mpz_clear(dsa_m), mpz_clear(dsa_r), mpz_clear(dsa_s);
@@ -383,6 +407,7 @@ void run_instance
 		gcry_mpi_release(q);
 		gcry_mpi_release(g);
 		gcry_mpi_release(y);
+		gcry_mpi_release(x);
 		delete dkg, delete dss, delete rbc, delete vtmf, delete aiou, delete aiou2;
 		exit(-1);
 	}
@@ -410,7 +435,8 @@ void run_instance
 		std::vector<gcry_mpi_t> qual, x_rvss_qual;
 		std::vector<std::string> capl; // canonicalized peer list
 		std::vector< std::vector<gcry_mpi_t> > c_ik;
-		if (!mpz_get_gcry_mpi(&h, dss->h))
+		h = gcry_mpi_new(2048);
+		if (!mpz_get_gcry_mpi(h, dss->h))
 		{
 			std::cerr << "P_" << whoami << ": mpz_get_gcry_mpi() failed for dss->h" << std::endl;
 			mpz_clear(dsa_m), mpz_clear(dsa_r), mpz_clear(dsa_s);
@@ -419,6 +445,7 @@ void run_instance
 			gcry_mpi_release(g);
 			gcry_mpi_release(y);
 			gcry_mpi_release(x);
+			gcry_mpi_release(h);
 			gcry_sexp_release(key);
 			delete dkg, delete dss, delete rbc, delete vtmf, delete aiou, delete aiou2;
 			exit(-1);
@@ -447,7 +474,8 @@ void run_instance
 			for (size_t k = 0; k <= dss->t; k++)
 			{
 				gcry_mpi_t tmp;
-				if (!mpz_get_gcry_mpi(&tmp, dss->dkg->x_rvss->C_ik[j][k]))
+				tmp = gcry_mpi_new(2048);
+				if (!mpz_get_gcry_mpi(tmp, dss->dkg->x_rvss->C_ik[j][k]))
 				{
 					std::cerr << "P_" << whoami << ": mpz_get_gcry_mpi() failed for dss->dkg->x_rvss->C_ik[j][k]" << std::endl;
 					mpz_clear(dsa_m), mpz_clear(dsa_r), mpz_clear(dsa_s);
@@ -469,6 +497,7 @@ void run_instance
 					for (size_t jj = 0; jj < c_ik.size(); jj++)
 						for (size_t kk = 0; kk < c_ik[jj].size(); kk++)
 							gcry_mpi_release(c_ik[jj][kk]);
+					gcry_mpi_release(tmp);
 					gcry_sexp_release(key);
 					delete dkg, delete dss, delete rbc, delete vtmf, delete aiou, delete aiou2;
 					exit(-1); 
@@ -476,7 +505,8 @@ void run_instance
 				c_ik[j].push_back(tmp);
 			}
 		}
-		if (!mpz_get_gcry_mpi(&x_i, dss->x_i))
+		x_i = gcry_mpi_new(2048);
+		if (!mpz_get_gcry_mpi(x_i, dss->x_i))
 		{
 			std::cerr << "P_" << whoami << ": mpz_get_gcry_mpi() failed for dss->x_i" << std::endl;
 			mpz_clear(dsa_m), mpz_clear(dsa_r), mpz_clear(dsa_s);
@@ -498,11 +528,13 @@ void run_instance
 			for (size_t j = 0; j < c_ik.size(); j++)
 				for (size_t k = 0; k < c_ik[j].size(); k++)
 					gcry_mpi_release(c_ik[j][k]);
+			gcry_mpi_release(x_i);
 			gcry_sexp_release(key);
 			delete dkg, delete dss, delete rbc, delete vtmf, delete aiou, delete aiou2;
 			exit(-1);
 		}
-		if (!mpz_get_gcry_mpi(&xprime_i, dss->xprime_i))
+		xprime_i = gcry_mpi_new(2048);
+		if (!mpz_get_gcry_mpi(xprime_i, dss->xprime_i))
 		{
 			std::cerr << "P_" << whoami << ": mpz_get_gcry_mpi() failed for dss->xprime_i" << std::endl;
 			mpz_clear(dsa_m), mpz_clear(dsa_r), mpz_clear(dsa_s);
@@ -525,6 +557,7 @@ void run_instance
 				for (size_t k = 0; k < c_ik[j].size(); k++)
 					gcry_mpi_release(c_ik[j][k]);
 			gcry_mpi_release(x_i);
+			gcry_mpi_release(xprime_i);
 			gcry_sexp_release(key);
 			delete dkg, delete dss, delete rbc, delete vtmf, delete aiou, delete aiou2;
 			exit(-1);
@@ -572,14 +605,13 @@ void run_instance
 		tmcg_byte_t buffer[1024];
 		gcry_mpi_t h;
 		size_t buflen = 0;
+		memset(buffer, 0, sizeof(buffer));
 		for (size_t i = 0; ((i < hash.size()) && (i < sizeof(buffer))); i++, buflen++)
 			buffer[i] = hash[i];
-		h = gcry_mpi_new(2048);
 		ret = gcry_mpi_scan(&h, GCRYMPI_FMT_USG, buffer, buflen, NULL);
 		if (ret)
 		{
 			std::cerr << "P_" << whoami << ": gcry_mpi_scan() failed for h" << std::endl;
-			gcry_mpi_release(h);
 			mpz_clear(dsa_m), mpz_clear(dsa_r), mpz_clear(dsa_s);
 			gcry_mpi_release(p);
 			gcry_mpi_release(q);
@@ -624,7 +656,8 @@ void run_instance
 		}
 		if (opt_verbose > 1)
 			std::cout << "P_" << whoami << ": log follows " << std::endl << err_log_sign.str();
-		if (!mpz_get_gcry_mpi(&r, dsa_r))
+		r = gcry_mpi_new(2048);
+		if (!mpz_get_gcry_mpi(r, dsa_r))
 		{
 			std::cerr << "P_" << whoami << ": mpz_get_gcry_mpi() failed for dsa_r" << std::endl;
 			mpz_clear(dsa_m), mpz_clear(dsa_r), mpz_clear(dsa_s);
@@ -633,11 +666,13 @@ void run_instance
 			gcry_mpi_release(g);
 			gcry_mpi_release(y);
 			gcry_mpi_release(x);
+			gcry_mpi_release(r);
 			gcry_sexp_release(key);
 			delete dkg, delete dss, delete rbc, delete vtmf, delete aiou, delete aiou2;
 			exit(-1);
 		}
-		if (!mpz_get_gcry_mpi(&s, dsa_s))
+		s = gcry_mpi_new(2048);
+		if (!mpz_get_gcry_mpi(s, dsa_s))
 		{
 			std::cerr << "P_" << whoami << ": mpz_get_gcry_mpi() failed for dsa_s" << std::endl;
 			mpz_clear(dsa_m), mpz_clear(dsa_r), mpz_clear(dsa_s);
@@ -647,6 +682,7 @@ void run_instance
 			gcry_mpi_release(y);
 			gcry_mpi_release(x);
 			gcry_mpi_release(r);
+			gcry_mpi_release(s);
 			gcry_sexp_release(key);
 			delete dkg, delete dss, delete rbc, delete vtmf, delete aiou, delete aiou2;
 			exit(-1);
@@ -680,13 +716,15 @@ void run_instance
 	gcry_mpi_release(y);
 	if (T > 0)
 	{
-		if (!mpz_get_gcry_mpi(&y, dkg->y)) // computed by DKG (cf. LibTMCG source code)
+		y = gcry_mpi_new(2048);
+		if (!mpz_get_gcry_mpi(y, dkg->y)) // computed by DKG (cf. LibTMCG source code)
 		{
 			std::cerr << "P_" << whoami << ": mpz_get_gcry_mpi() failed for dkg->y" << std::endl;
 			mpz_clear(dsa_m), mpz_clear(dsa_r), mpz_clear(dsa_s);
 			gcry_mpi_release(p);
 			gcry_mpi_release(q);
 			gcry_mpi_release(g);
+			gcry_mpi_release(y);
 			gcry_sexp_release(key);
 			delete dkg, delete dss, delete rbc, delete vtmf, delete aiou, delete aiou2;
 			exit(-1);
@@ -696,7 +734,8 @@ void run_instance
 		gcry_mpi_t h, n, t, i, qualsize, x_i, xprime_i;
 		std::vector<gcry_mpi_t> qual, v_i;
 		std::vector< std::vector<gcry_mpi_t> > c_ik;
-		if (!mpz_get_gcry_mpi(&h, dkg->h))
+		h = gcry_mpi_new(2048);
+		if (!mpz_get_gcry_mpi(h, dkg->h))
 		{
 			std::cerr << "P_" << whoami << ": mpz_get_gcry_mpi() failed for dkg->h" << std::endl;
 			mpz_clear(dsa_m), mpz_clear(dsa_r), mpz_clear(dsa_s);
@@ -704,6 +743,7 @@ void run_instance
 			gcry_mpi_release(q);
 			gcry_mpi_release(g);
 			gcry_mpi_release(y);
+			gcry_mpi_release(h);
 			gcry_sexp_release(key);
 			delete dkg, delete dss, delete rbc, delete vtmf, delete aiou, delete aiou2;
 			exit(-1);
@@ -720,7 +760,8 @@ void run_instance
 		v_i.resize(dkg->n);
 		for (size_t j = 0; j < v_i.size(); j++)
 		{
-			if (!mpz_get_gcry_mpi(&v_i[j], dkg->v_i[j]))
+			v_i[j] = gcry_mpi_new(2048);
+			if (!mpz_get_gcry_mpi(v_i[j], dkg->v_i[j]))
 			{
 				std::cerr << "P_" << whoami << ": mpz_get_gcry_mpi() failed for dkg->v_i[j]" << std::endl;
 				mpz_clear(dsa_m), mpz_clear(dsa_r), mpz_clear(dsa_s);
@@ -735,11 +776,14 @@ void run_instance
 				gcry_mpi_release(qualsize);
 				for (size_t jj = 0; jj < qual.size(); jj++)
 					gcry_mpi_release(qual[jj]);
-				for (size_t j = 0; j < v_i.size(); j++)
-					gcry_mpi_release(v_i[j]);
+				for (size_t jj = 0; jj < v_i.size(); jj++)
+					gcry_mpi_release(v_i[jj]);
 				for (size_t jj = 0; jj < c_ik.size(); jj++)
 					for (size_t kk = 0; kk < c_ik[jj].size(); kk++)
 						gcry_mpi_release(c_ik[jj][kk]);
+				gcry_mpi_release(v_i[j]);
+				gcry_sexp_release(key);
+				delete dkg, delete dss, delete rbc, delete vtmf, delete aiou, delete aiou2;
 				exit(-1); 
 			}
 		}
@@ -749,7 +793,8 @@ void run_instance
 			for (size_t k = 0; k <= dkg->t; k++)
 			{
 				gcry_mpi_t tmp;
-				if (!mpz_get_gcry_mpi(&tmp, dkg->C_ik[j][k]))
+				tmp = gcry_mpi_new(2048);
+				if (!mpz_get_gcry_mpi(tmp, dkg->C_ik[j][k]))
 				{
 					std::cerr << "P_" << whoami << ": mpz_get_gcry_mpi() failed for dkg->C_ik[j][k]" << std::endl;
 					mpz_clear(dsa_m), mpz_clear(dsa_r), mpz_clear(dsa_s);
@@ -764,17 +809,21 @@ void run_instance
 					gcry_mpi_release(qualsize);
 					for (size_t jj = 0; jj < qual.size(); jj++)
 						gcry_mpi_release(qual[jj]);
-					for (size_t j = 0; j < v_i.size(); j++)
-						gcry_mpi_release(v_i[j]);
+					for (size_t jj = 0; jj < v_i.size(); jj++)
+						gcry_mpi_release(v_i[jj]);
 					for (size_t jj = 0; jj < c_ik.size(); jj++)
 						for (size_t kk = 0; kk < c_ik[jj].size(); kk++)
 							gcry_mpi_release(c_ik[jj][kk]);
+					gcry_mpi_release(tmp);
+					gcry_sexp_release(key);
+					delete dkg, delete dss, delete rbc, delete vtmf, delete aiou, delete aiou2;
 					exit(-1); 
 				}
 				c_ik[j].push_back(tmp);
 			}
 		}
-		if (!mpz_get_gcry_mpi(&x_i, dkg->x_i))
+		x_i = gcry_mpi_new(2048);
+		if (!mpz_get_gcry_mpi(x_i, dkg->x_i))
 		{
 			std::cerr << "P_" << whoami << ": mpz_get_gcry_mpi() failed for dkg->x_i" << std::endl;
 			mpz_clear(dsa_m), mpz_clear(dsa_r), mpz_clear(dsa_s);
@@ -794,11 +843,13 @@ void run_instance
 			for (size_t j = 0; j < c_ik.size(); j++)
 				for (size_t k = 0; k < c_ik[j].size(); k++)
 					gcry_mpi_release(c_ik[j][k]);
+			gcry_mpi_release(x_i);
 			gcry_sexp_release(key);
 			delete dkg, delete dss, delete rbc, delete vtmf, delete aiou, delete aiou2;
 			exit(-1);
 		}
-		if (!mpz_get_gcry_mpi(&xprime_i, dkg->xprime_i))
+		xprime_i = gcry_mpi_new(2048);
+		if (!mpz_get_gcry_mpi(xprime_i, dkg->xprime_i))
 		{
 			std::cerr << "P_" << whoami << ": mpz_get_gcry_mpi() failed for dkg->xprime_i" << std::endl;
 			mpz_clear(dsa_m), mpz_clear(dsa_r), mpz_clear(dsa_s);
@@ -819,6 +870,7 @@ void run_instance
 				for (size_t k = 0; k < c_ik[j].size(); k++)
 					gcry_mpi_release(c_ik[j][k]);
 			gcry_mpi_release(x_i);
+			gcry_mpi_release(xprime_i);
 			gcry_sexp_release(key);
 			delete dkg, delete dss, delete rbc, delete vtmf, delete aiou, delete aiou2;
 			exit(-1);
@@ -856,14 +908,13 @@ void run_instance
 			tmcg_byte_t buffer[1024];
 			gcry_mpi_t h;
 			size_t buflen = 0;
+			memset(buffer, 0, sizeof(buffer));
 			for (size_t i = 0; ((i < hash.size()) && (i < sizeof(buffer))); i++, buflen++)
 				buffer[i] = hash[i];
-			h = gcry_mpi_new(2048);
 			ret = gcry_mpi_scan(&h, GCRYMPI_FMT_USG, buffer, buflen, NULL);
 			if (ret)
 			{
 				std::cerr << "P_" << whoami << ": gcry_mpi_scan() failed for h" << std::endl;
-				gcry_mpi_release(h);
 				mpz_clear(dsa_m), mpz_clear(dsa_r), mpz_clear(dsa_s);
 				gcry_mpi_release(p);
 				gcry_mpi_release(q);
@@ -902,18 +953,21 @@ void run_instance
 			}
 			if (opt_verbose > 1)
 				std::cout << "P_" << whoami << ": log follows " << std::endl << err_log_sign.str();
-			if (!mpz_get_gcry_mpi(&r, dsa_r))
+			r = gcry_mpi_new(2048);
+			if (!mpz_get_gcry_mpi(r, dsa_r))
 			{
 				std::cerr << "P_" << whoami << ": mpz_get_gcry_mpi() failed for dsa_r" << std::endl;
 				mpz_clear(dsa_m), mpz_clear(dsa_r), mpz_clear(dsa_s);
 				gcry_mpi_release(p);
 				gcry_mpi_release(q);
 				gcry_mpi_release(g);
+				gcry_mpi_release(r);
 				gcry_sexp_release(key);
 				delete dkg, delete dss, delete rbc, delete vtmf, delete aiou, delete aiou2;
 				exit(-1);
 			}
-			if (!mpz_get_gcry_mpi(&s, dsa_s))
+			s = gcry_mpi_new(2048);
+			if (!mpz_get_gcry_mpi(s, dsa_s))
 			{
 				std::cerr << "P_" << whoami << ": mpz_get_gcry_mpi() failed for dsa_s" << std::endl;
 				mpz_clear(dsa_m), mpz_clear(dsa_r), mpz_clear(dsa_s);
@@ -921,6 +975,7 @@ void run_instance
 				gcry_mpi_release(q);
 				gcry_mpi_release(g);
 				gcry_mpi_release(r);
+				gcry_mpi_release(s);
 				gcry_sexp_release(key);
 				delete dkg, delete dss, delete rbc, delete vtmf, delete aiou, delete aiou2;
 				exit(-1);
@@ -1560,11 +1615,19 @@ int main
 		"ALNb15Fcj550V40uidU5fxkgvlK2Orwbg|8|DKGPGTESTSUITEDKGPGTESTSU"
 		"ITEDKGPGTESTSUITEDKGPGTESTSUITEDKGPGTESTSUIU2|1v|1k|";
 #endif
+
+	// check command line arguments
+	if ((opt_hostname != NULL) && (opt_passwords == NULL))
+	{
+		std::cerr << "ERROR: option \"-P\" is necessary due to insecure network" << std::endl;
+		return -1;
+	}
 	if (peers.size() < 1)
 	{
 		std::cerr << "ERROR: no peers given as argument; usage: " << usage << std::endl;
 		return -1;
 	}
+
 	// canonicalize peer list and setup threshold values
 	std::sort(peers.begin(), peers.end());
 	std::vector<std::string>::iterator it = std::unique(peers.begin(), peers.end());
@@ -1576,31 +1639,46 @@ int main
 		std::cerr << "ERROR: too few or too many peers given" << std::endl;
 		return -1;
 	}
-	if (!init_libTMCG())
-	{
-		std::cerr << "ERROR: initialization of LibTMCG failed" << std::endl;
-		return -1;
-	}
-	if ((opt_hostname != NULL) && (opt_passwords == NULL))
-	{
-		std::cerr << "ERROR: option \"-P\" is necessary due to insecure network" << std::endl;
-		return -1;
-	}
-#ifdef DKGPG_TESTSUITE
-	userid = "TestGroup <testing@localhost>";
-	passphrase = "Test";
-#else
-	std::cout << "1. Please enter an OpenPGP-style user ID (name <email>): ";
-	std::getline(std::cin, userid);
-	std::cout << "2. Passphrase to protect your part of the private key (input will be displayed): ";
-	std::getline(std::cin, passphrase);
-#endif
 	if (opt_verbose)
 	{
 		std::cout << "INFO: canonicalized peer list = " << std::endl;
 		for (size_t i = 0; i < peers.size(); i++)
 			std::cout << peers[i] << std::endl;
 	}
+
+	// initialize LibTMCG
+	if (!init_libTMCG())
+	{
+		std::cerr << "ERROR: initialization of LibTMCG failed" << std::endl;
+		return -1;
+	}
+	if (opt_verbose)
+		std::cout << "INFO: using LibTMCG version " << version_libTMCG() << std::endl;
+
+	// read userid and passphrase
+#ifdef DKGPG_TESTSUITE
+	userid = "TestGroup <testing@localhost>";
+	passphrase = "Test";
+#else
+	std::cout << "1. Please enter an OpenPGP-style user ID (name <email>): ";
+	std::getline(std::cin, userid);
+	std::cin.clear();
+	std::string passphrase_check;
+	do
+	{
+		passphrase = "", passphrase_check = "";
+		if (!get_passphrase("2. Passphrase to protect your part of the private key", passphrase))
+			return -1;
+		if (!get_passphrase("Please repeat the given passphrase to continue", passphrase_check))
+			return -1;
+		if (passphrase != passphrase_check)
+			std::cerr << "WARNING: passphrase does not match; please try again" << std::endl;
+		else if (passphrase == "")
+			std::cerr << "WARNING: no key protection due to empty passphrase" << std::endl;
+	}
+	while (passphrase != passphrase_check);
+#endif
+
 #ifdef GNUNET
 	if (gnunet_opt_t_resilience != DKGPG_MAX_N)
 		T = gnunet_opt_t_resilience; // get value of T from GNUnet options
@@ -1750,6 +1828,8 @@ int main
 			mpz_clear(fips_hashalgo), mpz_clear(fips_dps), mpz_clear(fips_counter), mpz_clear(fips_index);
 			return -1;
 		}
+		// release
+		mpz_clear(fips_hashalgo), mpz_clear(fips_dps), mpz_clear(fips_counter), mpz_clear(fips_index);
 	}
 	// initialize cache
 	std::cout << "3. We need a lot of entropy to cache very strong randomness for key generation." << std::endl;
